@@ -5,7 +5,7 @@
 %%% Created :  8 Aug 2009 by Evgeniy Khramtsov <ekhramtsov@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2021 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2022 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -62,7 +62,6 @@
 -record(state,
 	{sock                        :: inet:socket() | fast_tls:tls_socket() | undefined,
 	 sock_mod = gen_tcp          :: gen_udp | gen_tcp | fast_tls,
-	 certfile                    :: iodata() | undefined,
 	 peer = {{0,0,0,0}, 0}       :: addr(),
 	 tref                        :: reference() | undefined,
 	 use_turn = false            :: boolean(),
@@ -73,7 +72,8 @@
 	 max_allocs = 10             :: non_neg_integer() | infinity,
 	 shaper = none               :: stun_shaper:shaper(),
 	 max_permissions = 10        :: non_neg_integer() | infinity,
-	 blacklist = []              :: turn:blacklist(),
+	 blacklist = []              :: turn:accesslist(),
+	 whitelist = []              :: turn:accesslist(),
 	 auth = user                 :: anonymous | user,
 	 nonces = treap:empty()      :: treap:treap(),
 	 realm = <<"">>              :: binary(),
@@ -107,11 +107,11 @@ udp_init(Sock, Opts) ->
 udp_recv(Sock, Addr, Port, Data, State) ->
     NewState = prepare_state(State, Sock, {Addr, Port}, gen_udp),
     case stun_codec:decode(Data, datagram) of
- 	{ok, Msg} ->
- 	    ?LOG_DEBUG(#{verbatim => {"Received:~n~s", [stun_codec:pp(Msg)]}}),
- 	    process(NewState, Msg);
- 	Err ->
-	    ?LOG_DEBUG("Cannot parse packet: ~p", [Err]),
+	{ok, Msg} ->
+	    ?LOG_DEBUG(#{verbatim => {"Received:~n~s", [stun_codec:pp(Msg)]}}),
+	    process(NewState, Msg);
+	{error, Reason} ->
+	    ?LOG_DEBUG("Cannot parse packet: ~s", [Reason]),
 	    NewState
     end.
 
@@ -125,8 +125,7 @@ init([Sock, Opts]) ->
 	    case get_sockmod(Opts, Sock) of
 		{ok, SockMod} ->
 		    State = prepare_state(Opts, Sock, Addr, SockMod),
-		    CertFile = get_certfile(Opts),
-		    case maybe_starttls(Sock, SockMod, CertFile) of
+		    case maybe_starttls(Sock, SockMod, Opts) of
 			{ok, NewSock} ->
 			    inet:setopts(Sock, [{active, once}]),
 			    TRef = erlang:start_timer(?TIMEOUT, self(), stop),
@@ -138,8 +137,8 @@ init([Sock, Opts]) ->
 		{error, Reason} ->
 		    {stop, Reason}
 	    end;
-	Err ->
-	    {stop, Err}
+	{error, Reason} ->
+	    {stop, Reason}
     end.
 
 session_established(Event, State) ->
@@ -160,8 +159,8 @@ handle_info({tcp, _Sock, TLSData}, StateName,
     case fast_tls:recv_data(NewState#state.sock, TLSData) of
 	{ok, Data} ->
 	    process_data(StateName, NewState, Data);
-	Err ->
-	    ?LOG_INFO("Connection failure: ~p", [Err]),
+	{error, Reason} ->
+	    ?LOG_INFO("Connection failure: ~s", [Reason]),
 	    {stop, normal, NewState}
     end;
 handle_info({tcp, _Sock, Data}, StateName, State) ->
@@ -243,12 +242,13 @@ process(#state{auth = user} = State,
 		<<"">> ->
 		    ?LOG_NOTICE("Failed long-term STUN/TURN authentication"),
 		    send(NewState, R);
-		Pass ->
+		Pass0 ->
+		    {Pass, IsExpired} = check_expired_tag(Pass0),
 		    case check_integrity(User, Realm, Msg, Pass) of
 			{true, Key} ->
 			    ?LOG_INFO("Accepting long-term STUN/TURN "
 				      "authentication"),
-			    process(NewState, Msg, Key);
+			    process(NewState, Msg, Key, IsExpired);
 			false ->
 			    ?LOG_NOTICE("Failed long-term STUN/TURN "
 					"authentication"),
@@ -287,8 +287,11 @@ process(State, Msg) when is_record(Msg, turn) ->
 process(State, _Msg) ->
     State.
 
+process(State, Msg, Secret) ->
+    process(State, Msg, Secret, false).
+
 process(State, #stun{class = request, unsupported = [_|_] = Unsupported} = Msg,
-	Secret) ->
+	Secret, _IsExpired) ->
     ?LOG_DEBUG("Rejecting request with unknown attribute(s): ~p",
 	       [Unsupported]),
     Resp = prepare_response(State, Msg),
@@ -297,7 +300,8 @@ process(State, #stun{class = request, unsupported = [_|_] = Unsupported} = Msg,
 		  'ERROR-CODE' = stun_codec:error(420)},
     send(State, R, Secret);
 process(State, #stun{class = request,
-		     method = ?STUN_METHOD_BINDING} = Msg, Secret) ->
+		     method = ?STUN_METHOD_BINDING} = Msg, Secret,
+	_IsExpired) ->
     Resp = prepare_response(State, Msg),
     AddrPort = unmap_v4_addr(State#state.peer),
     R = case stun_codec:version(Msg) of
@@ -311,14 +315,14 @@ process(State, #stun{class = request,
     run_hook(stun_query, State, Msg),
     send(State, R, Secret);
 process(#state{use_turn = false} = State,
-	#stun{class = request} = Msg, Secret) ->
+	#stun{class = request} = Msg, Secret, _IsExpired) ->
     ?LOG_NOTICE("Rejecting TURN request: TURN is disabled"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(405)},
     send(State, R, Secret);
 process(State, #stun{class = request,
 		     method = ?STUN_METHOD_ALLOCATE} = Msg,
-	Secret) ->
+	Secret, IsExpired) ->
     Resp = prepare_response(State, Msg),
     AddrPort = State#state.peer,
     SockMod = State#state.sock_mod,
@@ -326,16 +330,21 @@ process(State, #stun{class = request,
 	{ok, Pid} ->
 	    turn:route(Pid, Msg),
 	    State;
+	_ when IsExpired ->
+	    ?LOG_NOTICE("Rejecting request: credentials expired"),
+	    R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(401)},
+	    send(State, R);
 	_ ->
 	    Opts = [{sock, State#state.sock},
 		    {sock_mod, SockMod},
 		    {username, Msg#stun.'USERNAME'},
 		    {realm, State#state.realm},
 		    {key, Secret},
-		    {server, State#state.server_name},
+		    {server_name, State#state.server_name},
 		    {max_allocs, State#state.max_allocs},
 		    {max_permissions, State#state.max_permissions},
 		    {blacklist, State#state.blacklist},
+		    {whitelist, State#state.whitelist},
 		    {addr, AddrPort},
 		    {relay_ipv4_ip, State#state.relay_ipv4_ip},
 		    {relay_ipv6_ip, State#state.relay_ipv6_ip},
@@ -363,23 +372,26 @@ process(State, #stun{class = request,
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(438)},
 		    send(State, R);
-		Err ->
-		    ?LOG_ERROR("Cannot start TURN session: ~p", [Err]),
+		{error, Reason} ->
+		    ?LOG_ERROR("Cannot start TURN session: ~s", [Reason]),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(500)},
 		    send(State, R, Secret)
 	    end
     end;
 process(State, #stun{class = request,
-		     method = ?STUN_METHOD_REFRESH} = Msg, Secret) ->
+		     method = ?STUN_METHOD_REFRESH} = Msg, Secret,
+	_IsExpired) ->
     route_on_turn(State, Msg, Secret);
 process(State, #stun{class = request,
-		     method = ?STUN_METHOD_CREATE_PERMISSION} = Msg, Secret) ->
+		     method = ?STUN_METHOD_CREATE_PERMISSION} = Msg, Secret,
+	_IsExpired) ->
     route_on_turn(State, Msg, Secret);
 process(State, #stun{class = request,
-		    method = ?STUN_METHOD_CHANNEL_BIND} = Msg, Secret) ->
+		    method = ?STUN_METHOD_CHANNEL_BIND} = Msg, Secret,
+	_IsExpired) ->
     route_on_turn(State, Msg, Secret);
-process(State, #stun{class = request} = Msg, Secret) ->
+process(State, #stun{class = request} = Msg, Secret, _IsExpired) ->
     ?LOG_NOTICE("Rejecting request: Method not allowed"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error, 'ERROR-CODE' = stun_codec:error(405)},
@@ -529,6 +541,15 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 					 [B]),
 			      State
 		      end;
+		 ({turn_whitelist, B}, State) ->
+		      case lists:all(fun is_valid_subnet/1, B) of
+			  true ->
+			      State#state{whitelist = B};
+			  false ->
+			      ?LOG_ERROR("Wrong 'turn_whitelist' value: ~p",
+					 [B]),
+			      State
+		      end;
 		 ({shaper, S}, State)
 		    when S == none orelse (is_integer(S) andalso (S > 0)) ->
 		      State#state{shaper = stun_shaper:new(S)};
@@ -572,6 +593,9 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 		 ({ip, _}, State) -> State;
 		 ({backlog, _}, State) -> State;
 		 ({certfile, _}, State) -> State;
+		 ({dhfile, _}, State) -> State;
+		 ({ciphers, _}, State) -> State;
+		 ({protocol_options, _}, State) -> State;
 		 ({tls, _}, State) -> State;
 		 (tls, State) -> State;
 		 ({sock_peer_name, _}, State) -> State;
@@ -673,6 +697,11 @@ check_integrity(User, Realm, Msg, [Pass | T]) ->
 	    check_integrity(User, Realm, Msg, T)
     end.
 
+check_expired_tag({expired, Pass}) ->
+    {Pass, true};
+check_expired_tag(Pass) ->
+    {Pass, false}.
+
 unmap_v4_addr({{0, 0, 0, 0, 0, 16#FFFF, D7, D8}, Port}) ->
     {{D7 bsr 8, D7 band 255, D8 bsr 8, D8 band 255}, Port};
 unmap_v4_addr(AddrPort) ->
@@ -722,14 +751,6 @@ get_peername(Sock, Opts) ->
 	    inet:peername(Sock)
     end.
 
-get_certfile(Opts) ->
-    case catch iolist_to_binary(proplists:get_value(certfile, Opts)) of
-	Filename when is_binary(Filename), Filename /= <<"">> ->
-	    Filename;
-	_ ->
-	    undefined
-    end.
-
 -ifdef(USE_OLD_INET_BACKEND).
 -dialyzer({[no_match], [get_sockmod/2]}).
 is_tls_handshake(_Sock) ->
@@ -750,12 +771,28 @@ is_tls_handshake({_, _, {_, Socket}}) ->
     end.
 -endif.
 
-maybe_starttls(_Sock, fast_tls, undefined) ->
-    ?LOG_ERROR("Cannot start TLS connection: option 'certfile' is not set"),
-    {error, eprotonosupport};
-maybe_starttls(Sock, fast_tls, CertFile) ->
-    fast_tls:tcp_to_tls(Sock, [verify_none, {certfile, CertFile}]);
-maybe_starttls(Sock, gen_tcp, _CertFile) ->
+maybe_starttls(Sock, fast_tls, Opts) ->
+    case proplists:is_defined(certfile, Opts) of
+	true ->
+	    TLSOpts = lists:filter(
+			fun({certfile, _Val}) ->
+				true;
+			   ({dhfile, _Val}) ->
+				true;
+			   ({ciphers, _Val}) ->
+				true;
+			   ({protocol_options, _Val}) ->
+				true;
+			   (_Opt) ->
+				false
+			end, Opts),
+	    fast_tls:tcp_to_tls(Sock, [verify_none | TLSOpts]);
+	false ->
+	    ?LOG_ERROR("Cannot accept TLS connection: "
+		       "option 'certfile' is not set"),
+	    {error, eprotonosupport}
+    end;
+maybe_starttls(Sock, gen_tcp, _Opts) ->
     {ok, Sock}.
 
 prepare_response(State, Msg) ->
